import logging
import random
from pathlib import Path

import cv2
from albumentations.augmentations.geometric import longest_max_size

from ritm_annotation.data.base import ISDataset
from ritm_annotation.data.sample import DSample
from ritm_annotation.utils.exp_imports.default import *

logger = logging.getLogger(__name__)


def get_train_augmentator(model_cfg):
    crop_size = model_cfg.crop_size
    return Compose(
        [
            UniformRandomResize(scale_range=(0.75, 1.40)),
            HorizontalFlip(),
            PadIfNeeded(
                min_height=crop_size[0], min_width=crop_size[1], border_mode=0
            ),
            RandomCrop(*crop_size),
            RandomBrightnessContrast(
                brightness_limit=(-0.25, 0.25),
                contrast_limit=(-0.15, 0.4),
                p=0.75,
            ),
            RGBShift(
                r_shift_limit=10, g_shift_limit=10, b_shift_limit=10, p=0.75
            ),
        ],
        p=1.0,
    )


def get_val_augmentator(model_cfg):
    crop_size = model_cfg.crop_size
    return Compose(
        [
            PadIfNeeded(
                min_height=crop_size[0], min_width=crop_size[1], border_mode=0
            ),
            RandomCrop(*crop_size),
        ],
        p=1.0,
    )


def get_points_sampler(model_cfg):
    return MultiPointSampler(
        model_cfg.num_max_points,
        prob_gamma=0.8,
        merge_objects_prob=0.15,
        max_num_merged_objects=2,
    )


class AnnotationDataset(ISDataset):
    def __init__(
        self,
        images_path: Path,
        masks_path: Path,
        split="train",
        dry_run=False,
        # the idea here is to resize the image to speed up data ingestion and training  # noqa:E501
        max_bigger_dimension=None,
        **kwargs,
    ):
        super(AnnotationDataset, self).__init__(**kwargs)
        self.images_path = images_path
        self.masks_path = masks_path
        self.max_bigger_dimension = max_bigger_dimension
        self.dataset_samples = []
        if not dry_run:
            for item in masks_path.iterdir():
                image_file = images_path / item.name
                if not item.is_dir():
                    logger.warn(
                        _(
                            "AnnotationDataset: found impurities: {item}"
                        ).format(item=item)
                    )
                    continue
                if not (image_file.exists() and image_file.is_file()):
                    logger.warn(
                        _("Found mask for {item_name} but not image").format(
                            item_name=item.name
                        )
                    )
                    continue
                has_mask = False
                for mask_file in item.iterdir():
                    has_mask = True
                if has_mask:
                    self.dataset_samples.append(item.name)
        self.dataset_samples.sort()

        total_amount = len(self.dataset_samples)
        train_amount = int(total_amount * 0.8)
        val_amount = total_amount - train_amount
        if split == "train":
            self.dataset_samples = self.dataset_samples[:train_amount]
        elif split == "val":
            self.dataset_samples = self.dataset_samples[-val_amount:]
        else:
            raise ValueError(_("split must be either train or val"))

    def get_sample(self, index: int) -> DSample:
        item = self.dataset_samples[index]
        image_path = self.images_path / item
        masks_path = self.masks_path / item

        image = cv2.imread(str(image_path))
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        if self.max_bigger_dimension is not None:
            image = longest_max_size(
                image, self.max_bigger_dimension, cv2.INTER_LINEAR
            )
        (h, w, *_rest) = image.shape

        mask_path = random.choice(list(masks_path.iterdir()))
        gt_mask = cv2.imread(str(mask_path), 0)

        if self.max_bigger_dimension is not None:
            gt_mask = longest_max_size(
                gt_mask, self.max_bigger_dimension, cv2.INTER_NEAREST_EXACT
            )
        gt_mask[gt_mask > 0] = 1
        gt_mask = gt_mask.astype("int32")
        logger.debug(
            _("Processed item {index}: '{item}' (shape: ({w}, {h})").format(
                index=index, item=item, w=w, h=h
            )
        )
        return DSample(image, gt_mask, objects_ids=[1], sample_id=index)



import torch
import torch.nn as nn

GLUON_RESNET_TORCH_HUB = "rwightman/pytorch-pretrained-gluonresnet"


class BasicBlockV1b(nn.Module):
    expansion = 1

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        dilation=1,
        downsample=None,
        previous_dilation=1,
        norm_layer=nn.BatchNorm2d,
    ):
        super(BasicBlockV1b, self).__init__()
        self.conv1 = nn.Conv2d(
            inplanes,
            planes,
            kernel_size=3,
            stride=stride,
            padding=dilation,
            dilation=dilation,
            bias=False,
        )
        self.bn1 = norm_layer(planes)
        self.conv2 = nn.Conv2d(
            planes,
            planes,
            kernel_size=3,
            stride=1,
            padding=previous_dilation,
            dilation=previous_dilation,
            bias=False,
        )
        self.bn2 = norm_layer(planes)

        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out = out + residual
        out = self.relu(out)

        return out


class BottleneckV1b(nn.Module):
    expansion = 4

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        dilation=1,
        downsample=None,
        previous_dilation=1,
        norm_layer=nn.BatchNorm2d,
    ):
        super(BottleneckV1b, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = norm_layer(planes)

        self.conv2 = nn.Conv2d(
            planes,
            planes,
            kernel_size=3,
            stride=stride,
            padding=dilation,
            dilation=dilation,
            bias=False,
        )
        self.bn2 = norm_layer(planes)

        self.conv3 = nn.Conv2d(
            planes, planes * self.expansion, kernel_size=1, bias=False
        )
        self.bn3 = norm_layer(planes * self.expansion)

        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out = out + residual
        out = self.relu(out)

        return out


class ResNetV1b(nn.Module):
    """Pre-trained ResNetV1b Model, which produces the strides of 8 featuremaps at conv5.

    Parameters
    ----------
    block : Block
        Class for the residual block. Options are BasicBlockV1, BottleneckV1.
    layers : list of int
        Numbers of layers in each block
    classes : int, default 1000
        Number of classification classes.
    dilated : bool, default False
        Applying dilation strategy to pretrained ResNet yielding a stride-8 model,
        typically used in Semantic Segmentation.
    norm_layer : object
        Normalization layer used (default: :class:`nn.BatchNorm2d`)
    deep_stem : bool, default False
        Whether to replace the 7x7 conv1 with 3 3x3 convolution layers.
    avg_down : bool, default False
        Whether to use average pooling for projection skip connection between stages/downsample.
    final_drop : float, default 0.0
        Dropout ratio before the final classification layer.

    Reference:
        - He, Kaiming, et al. "Deep residual learning for image recognition."
        Proceedings of the IEEE conference on computer vision and pattern recognition. 2016.

        - Yu, Fisher, and Vladlen Koltun. "Multi-scale context aggregation by dilated convolutions."
    """  # noqa: E501

    def __init__(
        self,
        block,
        layers,
        classes=1000,
        dilated=True,
        deep_stem=False,
        stem_width=32,
        avg_down=False,
        final_drop=0.0,
        norm_layer=nn.BatchNorm2d,
    ):
        self.inplanes = stem_width * 2 if deep_stem else 64
        super(ResNetV1b, self).__init__()
        if not deep_stem:
            self.conv1 = nn.Conv2d(
                3, 64, kernel_size=7, stride=2, padding=3, bias=False
            )
        else:
            self.conv1 = nn.Sequential(
                nn.Conv2d(
                    3,
                    stem_width,
                    kernel_size=3,
                    stride=2,
                    padding=1,
                    bias=False,
                ),
                norm_layer(stem_width),
                nn.ReLU(True),
                nn.Conv2d(
                    stem_width,
                    stem_width,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=False,
                ),
                norm_layer(stem_width),
                nn.ReLU(True),
                nn.Conv2d(
                    stem_width,
                    2 * stem_width,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=False,
                ),
            )
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(True)
        self.maxpool = nn.MaxPool2d(3, stride=2, padding=1)
        self.layer1 = self._make_layer(
            block, 64, layers[0], avg_down=avg_down, norm_layer=norm_layer
        )
        self.layer2 = self._make_layer(
            block,
            128,
            layers[1],
            stride=2,
            avg_down=avg_down,
            norm_layer=norm_layer,
        )
        if dilated:
            self.layer3 = self._make_layer(
                block,
                256,
                layers[2],
                stride=1,
                dilation=2,
                avg_down=avg_down,
                norm_layer=norm_layer,
            )
            self.layer4 = self._make_layer(
                block,
                512,
                layers[3],
                stride=1,
                dilation=4,
                avg_down=avg_down,
                norm_layer=norm_layer,
            )
        else:
            self.layer3 = self._make_layer(
                block,
                256,
                layers[2],
                stride=2,
                avg_down=avg_down,
                norm_layer=norm_layer,
            )
            self.layer4 = self._make_layer(
                block,
                512,
                layers[3],
                stride=2,
                avg_down=avg_down,
                norm_layer=norm_layer,
            )
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.drop = None
        if final_drop > 0.0:
            self.drop = nn.Dropout(final_drop)
        self.fc = nn.Linear(512 * block.expansion, classes)

    def _make_layer(
        self,
        block,
        planes,
        blocks,
        stride=1,
        dilation=1,
        avg_down=False,
        norm_layer=nn.BatchNorm2d,
    ):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = []
            if avg_down:
                if dilation == 1:
                    downsample.append(
                        nn.AvgPool2d(
                            kernel_size=stride,
                            stride=stride,
                            ceil_mode=True,
                            count_include_pad=False,
                        )
                    )
                else:
                    downsample.append(
                        nn.AvgPool2d(
                            kernel_size=1,
                            stride=1,
                            ceil_mode=True,
                            count_include_pad=False,
                        )
                    )
                downsample.extend(
                    [
                        nn.Conv2d(
                            self.inplanes,
                            out_channels=planes * block.expansion,
                            kernel_size=1,
                            stride=1,
                            bias=False,
                        ),
                        norm_layer(planes * block.expansion),
                    ]
                )
                downsample = nn.Sequential(*downsample)
            else:
                downsample = nn.Sequential(
                    nn.Conv2d(
                        self.inplanes,
                        out_channels=planes * block.expansion,
                        kernel_size=1,
                        stride=stride,
                        bias=False,
                    ),
                    norm_layer(planes * block.expansion),
                )

        layers = []
        if dilation in (1, 2):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    stride,
                    dilation=1,
                    downsample=downsample,
                    previous_dilation=dilation,
                    norm_layer=norm_layer,
                )
            )
        elif dilation == 4:
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    stride,
                    dilation=2,
                    downsample=downsample,
                    previous_dilation=dilation,
                    norm_layer=norm_layer,
                )
            )
        else:
            raise RuntimeError(
                _("=> unknown dilation size: {dilation}").format(dilation)
            )

        self.inplanes = planes * block.expansion
        for __ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    dilation=dilation,
                    previous_dilation=dilation,
                    norm_layer=norm_layer,
                )
            )

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        if self.drop is not None:
            x = self.drop(x)
        x = self.fc(x)

        return x


def _safe_state_dict_filtering(orig_dict, model_dict_keys):
    filtered_orig_dict = {}
    for k, v in orig_dict.items():
        if k in model_dict_keys:
            filtered_orig_dict[k] = v
        else:
            print(_("[ERROR] Failed to load <{k}> in backbone").format(k=k))
    return filtered_orig_dict


def resnet34_v1b(pretrained=False, **kwargs):
    model = ResNetV1b(BasicBlockV1b, [3, 4, 6, 3], **kwargs)
    if pretrained:
        model_dict = model.state_dict()
        filtered_orig_dict = _safe_state_dict_filtering(
            torch.hub.load(
                GLUON_RESNET_TORCH_HUB, "gluon_resnet34_v1b", pretrained=True
            ).state_dict(),
            model_dict.keys(),
        )
        model_dict.update(filtered_orig_dict)
        model.load_state_dict(model_dict)
    return model


def resnet50_v1s(pretrained=False, **kwargs):
    model = ResNetV1b(
        BottleneckV1b, [3, 4, 6, 3], deep_stem=True, stem_width=64, **kwargs
    )
    if pretrained:
        model_dict = model.state_dict()
        filtered_orig_dict = _safe_state_dict_filtering(
            torch.hub.load(
                GLUON_RESNET_TORCH_HUB, "gluon_resnet50_v1s", pretrained=True
            ).state_dict(),
            model_dict.keys(),
        )
        model_dict.update(filtered_orig_dict)
        model.load_state_dict(model_dict)
    return model


def resnet101_v1s(pretrained=False, **kwargs):
    model = ResNetV1b(
        BottleneckV1b, [3, 4, 23, 3], deep_stem=True, stem_width=64, **kwargs
    )
    if pretrained:
        model_dict = model.state_dict()
        filtered_orig_dict = _safe_state_dict_filtering(
            torch.hub.load(
                GLUON_RESNET_TORCH_HUB, "gluon_resnet101_v1s", pretrained=True
            ).state_dict(),
            model_dict.keys(),
        )
        model_dict.update(filtered_orig_dict)
        model.load_state_dict(model_dict)
    return model


def resnet152_v1s(pretrained=False, **kwargs):
    model = ResNetV1b(
        BottleneckV1b, [3, 8, 36, 3], deep_stem=True, stem_width=64, **kwargs
    )
    if pretrained:
        model_dict = model.state_dict()
        filtered_orig_dict = _safe_state_dict_filtering(
            torch.hub.load(
                GLUON_RESNET_TORCH_HUB, "gluon_resnet152_v1s", pretrained=True
            ).state_dict(),
            model_dict.keys(),
        )
        model_dict.update(filtered_orig_dict)
        model.load_state_dict(model_dict)
    return model



from functools import lru_cache

import cv2
import numpy as np


def visualize_instances(
    imask,
    bg_color=255,
    boundaries_color=None,
    boundaries_width=1,
    boundaries_alpha=0.8,
):
    num_objects = imask.max() + 1
    palette = get_palette(num_objects)
    if bg_color is not None:
        palette[0] = bg_color

    result = palette[imask].astype(np.uint8)
    if boundaries_color is not None:
        boundaries_mask = get_boundaries(
            imask, boundaries_width=boundaries_width
        )
        tresult = result.astype(np.float32)
        tresult[boundaries_mask] = boundaries_color
        tresult = tresult * boundaries_alpha + (1 - boundaries_alpha) * result
        result = tresult.astype(np.uint8)

    return result


@lru_cache(maxsize=16)
def get_palette(num_cls):
    palette = np.zeros(3 * num_cls, dtype=np.int32)

    for j in range(0, num_cls):
        lab = j
        i = 0

        while lab > 0:
            palette[j * 3 + 0] |= ((lab >> 0) & 1) << (7 - i)
            palette[j * 3 + 1] |= ((lab >> 1) & 1) << (7 - i)
            palette[j * 3 + 2] |= ((lab >> 2) & 1) << (7 - i)
            i = i + 1
            lab >>= 3

    return palette.reshape((-1, 3))


def visualize_mask(mask, num_cls):
    palette = get_palette(num_cls)
    mask[mask == -1] = 0

    return palette[mask].astype(np.uint8)


def visualize_proposals(
    proposals_info, point_color=(255, 0, 0), point_radius=1
):
    proposal_map, colors, candidates = proposals_info

    proposal_map = draw_probmap(proposal_map)
    for x, y in candidates:
        proposal_map = cv2.circle(
            proposal_map, (y, x), point_radius, point_color, -1
        )

    return proposal_map


def draw_probmap(x):
    return cv2.applyColorMap((x * 255).astype(np.uint8), cv2.COLORMAP_HOT)


def draw_points(image, points, color, radius=3):
    image = image.copy()
    for p in points:
        if p[0] < 0:
            continue
        if len(p) == 3:
            pradius = {0: 8, 1: 6, 2: 4}[p[2]] if p[2] < 3 else 2
        else:
            pradius = radius
        image = cv2.circle(image, (int(p[1]), int(p[0])), pradius, color, -1)

    return image


def draw_instance_map(x, palette=None):
    num_colors = x.max() + 1
    if palette is None:
        palette = get_palette(num_colors)

    return palette[x].astype(np.uint8)


def blend_mask(image, mask, alpha=0.6):
    if mask.min() == -1:
        mask = mask.copy() + 1

    imap = draw_instance_map(mask)
    result = (image * (1 - alpha) + alpha * imap).astype(np.uint8)
    return result


def get_boundaries(instances_masks, boundaries_width=1):
    boundaries = np.zeros(
        (instances_masks.shape[0], instances_masks.shape[1]), dtype=np.bool
    )

    for obj_id in np.unique(instances_masks.flatten()):
        if obj_id == 0:
            continue

        obj_mask = instances_masks == obj_id
        kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (3, 3))
        inner_mask = cv2.erode(
            obj_mask.astype(np.uint8), kernel, iterations=boundaries_width
        ).astype(np.bool)

        obj_boundary = np.logical_xor(
            obj_mask, np.logical_and(inner_mask, obj_mask)
        )
        boundaries = np.logical_or(boundaries, obj_boundary)
    return boundaries


def draw_with_blend_and_clicks(
    img,
    mask=None,
    alpha=0.6,
    clicks_list=None,
    pos_color=(0, 255, 0),
    neg_color=(255, 0, 0),
    radius=4,
):
    result = img.copy()

    if mask is not None:
        palette = get_palette(np.max(mask) + 1)
        rgb_mask = palette[mask.astype(np.uint8)]

        mask_region = (mask > 0).astype(np.uint8)
        result = (
            result * (1 - mask_region[:, :, np.newaxis])
            + (1 - alpha) * mask_region[:, :, np.newaxis] * result
            + alpha * rgb_mask
        )
        result = result.astype(np.uint8)

        # result = (result * (1 - alpha) + alpha * rgb_mask).astype(np.uint8)

    if clicks_list is not None and len(clicks_list) > 0:
        pos_points = [
            click.coords for click in clicks_list if click.is_positive
        ]
        neg_points = [
            click.coords for click in clicks_list if not click.is_positive
        ]

        result = draw_points(result, pos_points, pos_color, radius=radius)
        result = draw_points(result, neg_points, neg_color, radius=radius)

    return result



import logging
from pathlib import Path

import torch

from ritm_annotation.inference.utils import (
    find_checkpoint,
    load_single_is_model,
)

logger = logging.getLogger(__name__)

COMMAND_DESCRIPTION = _("Show information about a model pth file")


def command(subparser):
    subparser.add_argument("model", type=Path)

    def handle(args):
        logger.debug(_("Loading model"))
        assert args.model.exists() and args.model.is_file(), _(
            "Model must exist and be a file"
        )
        checkpoint_path = find_checkpoint(args.model.parent, args.model.name)
        state_dict = torch.load(checkpoint_path, map_location="cpu")
        print(f"model {state_dict['config']['class']}")
        for k, v in state_dict["config"]["params"].items():
            print(f"config {k} {v['value']}")
        for k, v in state_dict["state_dict"].items():
            print(
                f"state {k} {str(v.dtype).replace('torch.', '')}[{tuple(v.shape)}]"  # noqa:E501
            )  # noqa:E501
        model = load_single_is_model(state_dict, torch.device("cpu"))
        for line in repr(model).split("\n"):
            print(f"arch {line.rstrip()}")

    return handle



from copy import deepcopy

import numpy as np
from albumentations import ReplayCompose

from ritm_annotation.data.transforms import remove_image_only_transforms
from ritm_annotation.utils.misc import get_labels_with_sizes


class DSample:
    def __init__(
        self,
        image,
        encoded_masks,
        objects=None,
        objects_ids=None,
        ignore_ids=None,
        sample_id=None,
    ):
        self.image = image
        self.sample_id = sample_id

        if len(encoded_masks.shape) == 2:
            encoded_masks = encoded_masks[:, :, np.newaxis]
        self._encoded_masks = encoded_masks
        self._ignored_regions = []

        if objects_ids is not None:
            if not objects_ids or not isinstance(objects_ids[0], tuple):
                assert encoded_masks.shape[2] == 1
                objects_ids = [(0, obj_id) for obj_id in objects_ids]

            self._objects = dict()
            for indx, obj_mapping in enumerate(objects_ids):
                self._objects[indx] = {
                    "parent": None,
                    "mapping": obj_mapping,
                    "children": [],
                }

            if ignore_ids:
                if isinstance(ignore_ids[0], tuple):
                    self._ignored_regions = ignore_ids
                else:
                    self._ignored_regions = [
                        (0, region_id) for region_id in ignore_ids
                    ]
        else:
            self._objects = deepcopy(objects)

        self._augmented = False
        self._soft_mask_aug = None
        self._original_data = (
            self.image,
            self._encoded_masks,
            deepcopy(self._objects),
        )

    def augment(self, augmentator):
        self.reset_augmentation()
        aug_output = augmentator(image=self.image, mask=self._encoded_masks)
        self.image = aug_output["image"]
        self._encoded_masks = aug_output["mask"]

        aug_replay = aug_output.get("replay", None)
        if aug_replay:
            assert len(self._ignored_regions) == 0
            mask_replay = remove_image_only_transforms(aug_replay)
            self._soft_mask_aug = ReplayCompose._restore_for_replay(
                mask_replay
            )

        self._compute_objects_areas()
        self.remove_small_objects(min_area=1)

        self._augmented = True

    def reset_augmentation(self):
        if not self._augmented:
            return
        orig_image, orig_masks, orig_objects = self._original_data
        self.image = orig_image
        self._encoded_masks = orig_masks
        self._objects = deepcopy(orig_objects)
        self._augmented = False
        self._soft_mask_aug = None

    def remove_small_objects(self, min_area):
        if self._objects and "area" not in list(self._objects.values())[0]:
            self._compute_objects_areas()

        for obj_id, obj_info in list(self._objects.items()):
            if obj_info["area"] < min_area:
                self._remove_object(obj_id)

    def get_object_mask(self, obj_id):
        layer_indx, mask_id = self._objects[obj_id]["mapping"]
        obj_mask = (self._encoded_masks[:, :, layer_indx] == mask_id).astype(
            np.int32
        )
        if self._ignored_regions:
            for layer_indx, mask_id in self._ignored_regions:
                ignore_mask = self._encoded_masks[:, :, layer_indx] == mask_id
                obj_mask[ignore_mask] = -1

        return obj_mask

    def get_soft_object_mask(self, obj_id):
        assert self._soft_mask_aug is not None
        original_encoded_masks = self._original_data[1]
        layer_indx, mask_id = self._objects[obj_id]["mapping"]
        obj_mask = (
            original_encoded_masks[:, :, layer_indx] == mask_id
        ).astype(np.float32)
        obj_mask = self._soft_mask_aug(
            image=obj_mask, mask=original_encoded_masks
        )["image"]
        return np.clip(obj_mask, 0, 1)

    def get_background_mask(self):
        return np.max(self._encoded_masks, axis=2) == 0

    @property
    def objects_ids(self):
        return list(self._objects.keys())

    @property
    def gt_mask(self):
        assert len(self._objects) == 1
        return self.get_object_mask(self.objects_ids[0])

    @property
    def root_objects(self):
        return [
            obj_id
            for obj_id, obj_info in self._objects.items()
            if obj_info["parent"] is None
        ]

    def _compute_objects_areas(self):
        inverse_index = {
            node["mapping"]: node_id for node_id, node in self._objects.items()
        }
        ignored_regions_keys = set(self._ignored_regions)

        for layer_indx in range(self._encoded_masks.shape[2]):
            objects_ids, objects_areas = get_labels_with_sizes(
                self._encoded_masks[:, :, layer_indx]
            )
            for obj_id, obj_area in zip(objects_ids, objects_areas):
                inv_key = (layer_indx, obj_id)
                if inv_key in ignored_regions_keys:
                    continue
                try:
                    self._objects[inverse_index[inv_key]]["area"] = obj_area
                    del inverse_index[inv_key]
                except KeyError:
                    layer = self._encoded_masks[:, :, layer_indx]
                    layer[layer == obj_id] = 0
                    self._encoded_masks[:, :, layer_indx] = layer

        for obj_id in inverse_index.values():
            self._objects[obj_id]["area"] = 0

    def _remove_object(self, obj_id):
        obj_info = self._objects[obj_id]
        obj_parent = obj_info["parent"]
        for child_id in obj_info["children"]:
            self._objects[child_id]["parent"] = obj_parent

        if obj_parent is not None:
            parent_children = self._objects[obj_parent]["children"]
            parent_children = [x for x in parent_children if x != obj_id]
            self._objects[obj_parent]["children"] = (
                parent_children + obj_info["children"]
            )

        del self._objects[obj_id]

    def __len__(self):
        return len(self._objects)



import ritm_annotation.utils.i18n  # F401

"""CLI interface for ritm_annotation project.

Be creative! do whatever you want!

- Install click or typer and create a CLI app
- Use builtin argparse
- Start a web application
- Import things from your .base module
"""

import logging
import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from pathlib import Path

from ritm_annotation.utils.misc import load_module

logger = logging.getLogger(__name__)


def add_subcommand(subparsers, name: str, submodule):
    subparser = subparsers.add_parser(name, help=submodule.COMMAND_DESCRIPTION)
    common_flags(subparser)
    handler = submodule.command(subparser)
    subparser.set_defaults(fn=handler)


def common_flags(parser):
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help=_("Give more details about what is happening"),
    )  # noqa: E501
    parser.add_argument(
        "-V",
        "--version",
        dest="is_show_version",
        action="store_true",
        help=_("Print version and exit"),
    )  # noqa: E501


def main():  # pragma: no cover
    """
    The main function executes on commands:
    `python -m ritm_annotation` and `$ ritm_annotation `.

    This is your program's entry point.

    You can change this function to do whatever you want.
    Examples:
        * Run a test suite
        * Run a server
        * Do some other stuff
        * Run a command line application (Click, Typer, ArgParse)
        * List all available tasks
        * Run an application (Flask, FastAPI, Django, etc.)
    """
    logging.basicConfig()
    parser = ArgumentParser(
        prog="ritm_annotation", formatter_class=ArgumentDefaultsHelpFormatter
    )
    common_flags(parser)
    subparsers = parser.add_subparsers()

    for module in Path(__file__).parent.glob("*/__init__.py"):
        if str(module).find("pycache") > 0:
            continue
        module_name = module.parent.name
        subcommand_module = load_module(
            module, module_name=f"ritm_annotation.cli.{module_name}"
        )
        add_subcommand(subparsers, module_name, subcommand_module)

    args = parser.parse_args()

    if args.verbose:
        logging.root.setLevel(logging.DEBUG)

    version = open(str(Path(__file__).parent.parent / "VERSION"), "r").read()
    if args.is_show_version:
        print(version)
        exit(0)
    logger.debug(f"{_('Starting')} ritm_annotation v{version}")

    fn = args.__dict__.get("fn")
    args.__dict__["fn"] = None
    if fn is not None:
        fn(args)
    else:
        parser.parse_args([*sys.argv[1:], "--help"])



# flake8: noqa

from ritm_annotation.data.compose import (
    ComposeDataset,
    ProportionalComposeDataset,
)

from .ade20k import ADE20kDataset
from .berkeley import BerkeleyDataset
from .coco import CocoDataset
from .coco_lvis import CocoLvisDataset
from .davis import DavisDataset
from .grabcut import GrabCutDataset
from .images_dir import ImagesDirDataset
from .lvis import LvisDataset
from .openimages import OpenImagesDataset
from .pascalvoc import PascalVocDataset
from .sbd import SBDDataset, SBDEvaluationDataset



import logging
from typing import Dict

from easydict import EasyDict as edict

logger = logging.getLogger(__name__)


def load_cfg_from_env(cfg: edict, env: Dict[str, str]):
    for k, v in env.items():
        if k.startswith("RITM_"):
            cfgkey = k.replace("RITM_", "").replace("__", ".")
            logger.warning(
                _(
                    "Changing configuration entry from environment variable: {k}={v}"
                ).format(
                    k=cfgkey, v=v
                )  # noqa:E501
            )  # noqa: E501
            *parts, last = cfgkey.split(".")
            this_cfg = cfg
            for part in parts:
                if this_cfg.get(part) is None:
                    this_cfg[part] = edict()
                this_cfg = this_cfg[part]
            this_cfg[last] = v
    return cfg



import logging
import os
import shutil
import sys
from datetime import datetime
from pathlib import Path

import torch
import yaml  # type: ignore
from easydict import EasyDict as edict

from .distributed import get_world_size, synchronize

# from .log import add_logging

logger = logging.getLogger(__name__)


def init_experiment(args, model_name):
    model_path = Path(args.model_path)
    ftree = get_model_family_tree(model_path, model_name=model_name)

    if ftree is None:
        print(
            _(
                'Models can only be located in the "models" directory in the root of the repository'
            )  # noqa:E501
        )
        sys.exit(1)

    cfg = load_config(model_path)
    update_config(cfg, args)

    cfg.distributed = args.distributed
    cfg.local_rank = args.local_rank
    if cfg.distributed:
        torch.distributed.init_process_group(
            backend="nccl", init_method="env://"
        )
        if args.workers > 0:
            torch.multiprocessing.set_start_method("forkserver", force=True)

    experiments_path = Path(args.experiment_path)
    exp_parent_path = experiments_path / "/".join(ftree)
    exp_parent_path.mkdir(parents=True, exist_ok=True)

    if cfg.resume_exp:
        exp_path = find_resume_exp(exp_parent_path, cfg.resume_exp)
    else:
        last_exp_indx = find_last_exp_indx(exp_parent_path)
        exp_name = f"{last_exp_indx:03d}"
        if cfg.exp_name:
            exp_name += "_" + cfg.exp_name
        exp_path = exp_parent_path / exp_name
        synchronize()
        if cfg.local_rank == 0:
            exp_path.mkdir(parents=True)

    cfg.EXP_PATH = exp_path
    cfg.CHECKPOINTS_PATH = exp_path / "checkpoints"
    cfg.VIS_PATH = exp_path / "vis"
    cfg.LOGS_PATH = exp_path / "logs"

    if cfg.local_rank == 0:
        cfg.LOGS_PATH.mkdir(exist_ok=True)
        cfg.CHECKPOINTS_PATH.mkdir(exist_ok=True)
        cfg.VIS_PATH.mkdir(exist_ok=True)

        dst_script_path = exp_path / (
            model_path.stem
            + datetime.strftime(datetime.today(), "_%Y-%m-%d-%H-%M-%S.py")
        )
        if args.temp_model_path:
            shutil.copy(args.temp_model_path, dst_script_path)
            os.remove(args.temp_model_path)
        else:
            shutil.copy(model_path, dst_script_path)

    synchronize()

    if cfg.gpus != "":
        gpu_ids = [int(id) for id in cfg.gpus.split(",")]
    else:
        gpu_ids = list(range(max(cfg.ngpus, get_world_size())))
        cfg.gpus = ",".join([str(id) for id in gpu_ids])

    cfg.gpu_ids = gpu_ids
    cfg.ngpus = len(gpu_ids)
    cfg.multi_gpu = cfg.ngpus > 1

    if cfg.distributed:
        cfg.device = torch.device("cuda")
        cfg.gpu_ids = [cfg.gpu_ids[cfg.local_rank]]
        torch.cuda.set_device(cfg.gpu_ids[0])
    else:
        if cfg.multi_gpu:
            os.environ["CUDA_VISIBLE_DEVICES"] = cfg.gpus
            ngpus = torch.cuda.device_count()
            assert ngpus == cfg.ngpus
        cfg.device = torch.device(f"cuda:{cfg.gpu_ids[0]}")

    if cfg.local_rank == 0:
        # add_logging(cfg.LOGS_PATH, prefix="train_")
        logger.info(_("Number of GPUs: {n}").format(n=cfg.ngpus))
        if cfg.distributed:
            logger.info(_("Multi-Process Multi-GPU Distributed Training"))

    return cfg


def get_model_family_tree(
    model_path, terminate_name="models", model_name=None
):
    if model_name is None:
        model_name = model_path.stem
    family_tree = [model_name]
    for x in model_path.parents:
        if x.stem == terminate_name:
            break
        family_tree.append(x.stem)
    else:
        return None

    return family_tree[::-1]


def find_last_exp_indx(exp_parent_path):
    indx = 0
    for x in exp_parent_path.iterdir():
        if not x.is_dir():
            continue

        exp_name = x.stem
        if exp_name[:3].isnumeric():
            indx = max(indx, int(exp_name[:3]) + 1)

    return indx


def find_resume_exp(exp_parent_path, exp_pattern):
    candidates = sorted(exp_parent_path.glob(f"{exp_pattern}*"))
    if len(candidates) == 0:
        logger.warning(
            _(
                'No experiments could be found that satisfies the pattern = "{exp_pattern}"'
            ).format(pattern=f"*{exp_pattern}")
        )
        logger.info(
            _("Candidates: {candidates}").format(
                candidates=" ".join(
                    [e.name for e in exp_parent_path.iterdir()]
                )
            )
        )
        sys.exit(1)
    elif len(candidates) > 1:
        print(_("More than one experiment found:"))
        for x in candidates:
            print(x)
        sys.exit(1)
    else:
        exp_path = candidates[0]
        print(
            _('Continue with experiment "{exp_path}"').format(
                exp_path=exp_path
            )
        )

    return exp_path


def update_config(cfg, args):
    for param_name, value in vars(args).items():
        if param_name.lower() in cfg or param_name.upper() in cfg:
            continue
        cfg[param_name] = value


def load_config(model_path):
    model_name = model_path.stem
    config_path = model_path.parent / (model_name + ".yml")

    if config_path.exists():
        cfg = load_config_file(config_path)
    else:
        cfg = dict()

    cwd = Path.cwd()
    config_parent = config_path.parent.absolute()
    while len(config_parent.parents) > 0:
        config_path = config_parent / "config.yml"

        if config_path.exists():
            local_config = load_config_file(config_path, model_name=model_name)
            cfg.update({k: v for k, v in local_config.items() if k not in cfg})

        if config_parent.absolute() == cwd:
            break
        config_parent = config_parent.parent

    return edict(cfg)


def load_config_file(config_path, model_name=None, return_edict=False):
    with open(config_path, "r") as f:
        cfg = yaml.safe_load(f)

    if "SUBCONFIGS" in cfg:
        if model_name is not None and model_name in cfg["SUBCONFIGS"]:
            cfg.update(cfg["SUBCONFIGS"][model_name])
        del cfg["SUBCONFIGS"]

    return edict(cfg) if return_edict else cfg



import math
from typing import List

import numpy as np
import torch

from ritm_annotation.inference.clicker import Click

from .base import BaseTransform


class Crops(BaseTransform):
    def __init__(self, crop_size=(320, 480), min_overlap=0.2):
        super().__init__()
        self.crop_height, self.crop_width = crop_size
        self.min_overlap = min_overlap

        self.x_offsets = None
        self.y_offsets = None
        self._counts = None

    def transform(self, image_nd, clicks_lists: List[List[Click]]):
        assert image_nd.shape[0] == 1 and len(clicks_lists) == 1
        image_height, image_width = image_nd.shape[2:4]
        self._counts = None

        if image_height < self.crop_height or image_width < self.crop_width:
            return image_nd, clicks_lists

        self.x_offsets = get_offsets(
            image_width, self.crop_width, self.min_overlap
        )
        self.y_offsets = get_offsets(
            image_height, self.crop_height, self.min_overlap
        )
        self._counts = np.zeros((image_height, image_width))

        image_crops = []
        for dy in self.y_offsets:
            for dx in self.x_offsets:
                self._counts[
                    dy : dy + self.crop_height, dx : dx + self.crop_width
                ] += 1
                image_crop = image_nd[
                    :, :, dy : dy + self.crop_height, dx : dx + self.crop_width
                ]
                image_crops.append(image_crop)
        image_crops = torch.cat(image_crops, dim=0)  # type: ignore
        self._counts = torch.tensor(
            self._counts, device=image_nd.device, dtype=torch.float32
        )

        clicks_list = clicks_lists[0]
        clicks_lists = []
        for dy in self.y_offsets:
            for dx in self.x_offsets:
                crop_clicks = [
                    x.copy(coords=(x.coords[0] - dy, x.coords[1] - dx))
                    for x in clicks_list
                ]
                clicks_lists.append(crop_clicks)

        return image_crops, clicks_lists

    def inv_transform(self, prob_map):
        if self._counts is None:
            return prob_map

        new_prob_map = torch.zeros(
            (1, 1, *self._counts.shape),
            dtype=prob_map.dtype,
            device=prob_map.device,
        )

        crop_indx = 0
        for dy in self.y_offsets:
            for dx in self.x_offsets:
                new_prob_map[
                    0, 0, dy : dy + self.crop_height, dx : dx + self.crop_width
                ] += prob_map[crop_indx, 0]
                crop_indx += 1
        new_prob_map = torch.div(new_prob_map, self._counts)

        return new_prob_map

    def get_state(self):
        return self.x_offsets, self.y_offsets, self._counts

    def set_state(self, state):
        self.x_offsets, self.y_offsets, self._counts = state

    def reset(self):
        self.x_offsets = None
        self.y_offsets = None
        self._counts = None


def get_offsets(length, crop_size, min_overlap_ratio=0.2):
    if length == crop_size:
        return [0]

    N = (length / crop_size - min_overlap_ratio) / (1 - min_overlap_ratio)
    N = math.ceil(N)

    overlap_ratio = (N - length / crop_size) / (N - 1)
    overlap_width = int(crop_size * overlap_ratio)

    offsets = [0]
    for i in range(1, N):
        new_offset = offsets[-1] + crop_size - overlap_width
        if new_offset + crop_size > length:
            new_offset = length - crop_size

        offsets.append(new_offset)

    return offsets



import numpy as np
import torch

from ritm_annotation.utils import misc


class TrainMetric(object):
    def __init__(self, pred_outputs, gt_outputs):
        self.pred_outputs = pred_outputs
        self.gt_outputs = gt_outputs

    def update(self, *args, **kwargs):
        raise NotImplementedError

    def get_epoch_value(self):
        raise NotImplementedError

    def reset_epoch_stats(self):
        raise NotImplementedError

    def log_states(self, sw, tag_prefix, global_step):
        pass

    @property
    def name(self):
        return type(self).__name__


class AdaptiveIoU(TrainMetric):
    def __init__(
        self,
        init_thresh=0.4,
        thresh_step=0.025,
        thresh_beta=0.99,
        iou_beta=0.9,
        ignore_label=-1,
        from_logits=True,
        pred_output="instances",
        gt_output="instances",
    ):
        super().__init__(pred_outputs=(pred_output,), gt_outputs=(gt_output,))
        self._ignore_label = ignore_label
        self._from_logits = from_logits
        self._iou_thresh = init_thresh
        self._thresh_step = thresh_step
        self._thresh_beta = thresh_beta
        self._iou_beta = iou_beta
        self._ema_iou = 0.0
        self._epoch_iou_sum = 0.0
        self._epoch_batch_count = 0

    def update(self, pred, gt):
        gt_mask = gt > 0.5
        if self._from_logits:
            pred = torch.sigmoid(pred)

        gt_mask_area = torch.sum(gt_mask, dim=(1, 2)).detach().cpu().numpy()
        if np.all(gt_mask_area == 0):
            return

        ignore_mask = gt == self._ignore_label
        max_iou = _compute_iou(
            pred > self._iou_thresh, gt_mask, ignore_mask
        ).mean()
        best_thresh = self._iou_thresh
        for t in [
            best_thresh - self._thresh_step,
            best_thresh + self._thresh_step,
        ]:
            temp_iou = _compute_iou(pred > t, gt_mask, ignore_mask).mean()
            if temp_iou > max_iou:
                max_iou = temp_iou
                best_thresh = t

        self._iou_thresh = (
            self._thresh_beta * self._iou_thresh
            + (1 - self._thresh_beta) * best_thresh
        )
        self._ema_iou = (
            self._iou_beta * self._ema_iou + (1 - self._iou_beta) * max_iou
        )
        self._epoch_iou_sum += max_iou
        self._epoch_batch_count += 1

    def get_epoch_value(self):
        if self._epoch_batch_count > 0:
            return self._epoch_iou_sum / self._epoch_batch_count
        else:
            return 0.0

    def reset_epoch_stats(self):
        self._epoch_iou_sum = 0.0
        self._epoch_batch_count = 0

    def log_states(self, sw, tag_prefix, global_step):
        sw.add_scalar(
            tag=tag_prefix + "_ema_iou",
            value=self._ema_iou,
            global_step=global_step,
        )
        sw.add_scalar(
            tag=tag_prefix + "_iou_thresh",
            value=self._iou_thresh,
            global_step=global_step,
        )

    @property
    def iou_thresh(self):
        return self._iou_thresh


def _compute_iou(pred_mask, gt_mask, ignore_mask=None, keep_ignore=False):
    if ignore_mask is not None:
        pred_mask = torch.where(
            ignore_mask, torch.zeros_like(pred_mask), pred_mask
        )

    reduction_dims = misc.get_dims_with_exclusion(gt_mask.dim(), 0)
    union = (
        torch.mean((pred_mask | gt_mask).float(), dim=reduction_dims)
        .detach()
        .cpu()
        .numpy()
    )
    intersection = (
        torch.mean((pred_mask & gt_mask).float(), dim=reduction_dims)
        .detach()
        .cpu()
        .numpy()
    )
    nonzero = union > 0

    iou = intersection[nonzero] / union[nonzero]
    if not keep_ignore:
        return iou
    else:
        result = np.full_like(intersection, -1)
        result[nonzero] = iou
        return result



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from ritm_annotation.utils import misc


class NormalizedFocalLossSigmoid(nn.Module):
    def __init__(
        self,
        axis=-1,
        alpha=0.25,
        gamma=2,
        max_mult=-1,
        eps=1e-12,
        from_sigmoid=False,
        detach_delimeter=True,
        batch_axis=0,
        weight=None,
        size_average=True,
        ignore_label=-1,
    ):
        super(NormalizedFocalLossSigmoid, self).__init__()
        self._axis = axis
        self._alpha = alpha
        self._gamma = gamma
        self._ignore_label = ignore_label
        self._weight = weight if weight is not None else 1.0
        self._batch_axis = batch_axis

        self._from_logits = from_sigmoid
        self._eps = eps
        self._size_average = size_average
        self._detach_delimeter = detach_delimeter
        self._max_mult = max_mult
        self._k_sum = 0
        self._m_max = 0

    def forward(self, pred, label):
        one_hot = label > 0.5
        sample_weight = label != self._ignore_label

        if not self._from_logits:
            pred = torch.sigmoid(pred)

        alpha = torch.where(
            one_hot,
            self._alpha * sample_weight,
            (1 - self._alpha) * sample_weight,
        )
        pt = torch.where(
            sample_weight, 1.0 - torch.abs(label - pred), torch.ones_like(pred)
        )

        beta = (1 - pt) ** self._gamma

        sw_sum = torch.sum(sample_weight, dim=(-2, -1), keepdim=True)
        beta_sum = torch.sum(beta, dim=(-2, -1), keepdim=True)
        mult = sw_sum / (beta_sum + self._eps)
        if self._detach_delimeter:
            mult = mult.detach()
        beta = beta * mult
        if self._max_mult > 0:
            beta = torch.clamp_max(beta, self._max_mult)

        with torch.no_grad():
            ignore_area = (
                torch.sum(
                    label == self._ignore_label,
                    dim=tuple(range(1, label.dim())),
                )
                .cpu()
                .numpy()
            )
            sample_mult = (
                torch.mean(mult, dim=tuple(range(1, mult.dim()))).cpu().numpy()
            )
            if np.any(ignore_area == 0):
                self._k_sum = (
                    0.9 * self._k_sum
                    + 0.1 * sample_mult[ignore_area == 0].mean()
                )

                beta_pmax, _ = torch.flatten(beta, start_dim=1).max(dim=1)
                beta_pmax = beta_pmax.mean().item()
                self._m_max = 0.8 * self._m_max + 0.2 * beta_pmax

        loss = (
            -alpha
            * beta
            * torch.log(
                torch.min(
                    pt + self._eps,
                    torch.ones(1, dtype=torch.float).to(pt.device),
                )
            )
        )
        loss = self._weight * (loss * sample_weight)

        if self._size_average:
            bsum = torch.sum(
                sample_weight,
                dim=misc.get_dims_with_exclusion(
                    sample_weight.dim(), self._batch_axis
                ),
            )
            loss = torch.sum(
                loss,
                dim=misc.get_dims_with_exclusion(loss.dim(), self._batch_axis),
            ) / (bsum + self._eps)
        else:
            loss = torch.sum(
                loss,
                dim=misc.get_dims_with_exclusion(loss.dim(), self._batch_axis),
            )

        return loss

    def log_states(self, sw, name, global_step):
        sw.add_scalar(
            tag=name + "_k", value=self._k_sum, global_step=global_step
        )
        sw.add_scalar(
            tag=name + "_m", value=self._m_max, global_step=global_step
        )


class FocalLoss(nn.Module):
    def __init__(
        self,
        axis=-1,
        alpha=0.25,
        gamma=2,
        from_logits=False,
        batch_axis=0,
        weight=None,
        num_class=None,
        eps=1e-9,
        size_average=True,
        scale=1.0,
        ignore_label=-1,
    ):
        super(FocalLoss, self).__init__()
        self._axis = axis
        self._alpha = alpha
        self._gamma = gamma
        self._ignore_label = ignore_label
        self._weight = weight if weight is not None else 1.0
        self._batch_axis = batch_axis

        self._scale = scale
        self._num_class = num_class
        self._from_logits = from_logits
        self._eps = eps
        self._size_average = size_average

    def forward(self, pred, label, sample_weight=None):
        one_hot = label > 0.5
        sample_weight = label != self._ignore_label

        if not self._from_logits:
            pred = torch.sigmoid(pred)

        alpha = torch.where(
            one_hot,
            self._alpha * sample_weight,
            (1 - self._alpha) * sample_weight,
        )
        pt = torch.where(
            sample_weight, 1.0 - torch.abs(label - pred), torch.ones_like(pred)
        )

        beta = (1 - pt) ** self._gamma

        loss = (
            -alpha
            * beta
            * torch.log(
                torch.min(
                    pt + self._eps,
                    torch.ones(1, dtype=torch.float).to(pt.device),
                )
            )
        )
        loss = self._weight * (loss * sample_weight)

        if self._size_average:
            tsum = torch.sum(
                sample_weight,
                dim=misc.get_dims_with_exclusion(
                    label.dim(), self._batch_axis
                ),
            )
            loss = torch.sum(
                loss,
                dim=misc.get_dims_with_exclusion(loss.dim(), self._batch_axis),
            ) / (tsum + self._eps)
        else:
            loss = torch.sum(
                loss,
                dim=misc.get_dims_with_exclusion(loss.dim(), self._batch_axis),
            )

        return self._scale * loss


class SoftIoU(nn.Module):
    def __init__(self, from_sigmoid=False, ignore_label=-1):
        super().__init__()
        self._from_sigmoid = from_sigmoid
        self._ignore_label = ignore_label

    def forward(self, pred, label):
        label = label.view(pred.size())
        sample_weight = label != self._ignore_label

        if not self._from_sigmoid:
            pred = torch.sigmoid(pred)

        loss = 1.0 - torch.sum(pred * label * sample_weight, dim=(1, 2, 3)) / (
            torch.sum(torch.max(pred, label) * sample_weight, dim=(1, 2, 3))
            + 1e-8
        )

        return loss


class SigmoidBinaryCrossEntropyLoss(nn.Module):
    def __init__(
        self, from_sigmoid=False, weight=None, batch_axis=0, ignore_label=-1
    ):
        super(SigmoidBinaryCrossEntropyLoss, self).__init__()
        self._from_sigmoid = from_sigmoid
        self._ignore_label = ignore_label
        self._weight = weight if weight is not None else 1.0
        self._batch_axis = batch_axis

    def forward(self, pred, label):
        label = label.view(pred.size())
        sample_weight = label != self._ignore_label
        label = torch.where(sample_weight, label, torch.zeros_like(label))

        if not self._from_sigmoid:
            loss = (
                torch.relu(pred) - pred * label + F.softplus(-torch.abs(pred))
            )
        else:
            eps = 1e-12
            loss = -(
                torch.log(pred + eps) * label
                + torch.log(1.0 - pred + eps) * (1.0 - label)
            )

        loss = self._weight * (loss * sample_weight)
        return torch.mean(
            loss,
            dim=misc.get_dims_with_exclusion(loss.dim(), self._batch_axis),
        )



from pathlib import Path

import cv2
import numpy as np

from ritm_annotation.data.base import ISDataset
from ritm_annotation.data.sample import DSample


class DavisDataset(ISDataset):
    def __init__(
        self,
        dataset_path,
        images_dir_name="img",
        masks_dir_name="gt",
        **kwargs
    ):
        super(DavisDataset, self).__init__(**kwargs)

        self.dataset_path = Path(dataset_path)
        self._images_path = self.dataset_path / images_dir_name
        self._insts_path = self.dataset_path / masks_dir_name

        self.dataset_samples = [
            x.name for x in sorted(self._images_path.glob("*.*"))
        ]
        self._masks_paths = {x.stem: x for x in self._insts_path.glob("*.*")}

    def get_sample(self, index) -> DSample:
        image_name = self.dataset_samples[index]
        image_path = str(self._images_path / image_name)
        mask_path = str(self._masks_paths[image_name.split(".")[0]])

        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        instances_mask = np.max(cv2.imread(mask_path).astype(np.int32), axis=2)
        instances_mask[instances_mask > 0] = 1

        return DSample(image, instances_mask, objects_ids=[1], sample_id=index)
